<?php namespace EwTools\Models;

use EwTools\Support\Fluent;
use EwTools\Support\Str;
use EwTools\Support\Collection;
use WP_Post;

/**
 * Page
 *
 * @link    http://infomaker.se
 * @package EwTools\Models
 * @since   EwTools\Models\Page 0.1
 */
class Page extends Fluent {
    
    protected static $post_type = 'page';
    
    /**
     * Defaults for fetching pages in Wordpress
     *
     * @var array
     */
    protected static $default_options = [
        'sort_order'   => 'ASC',
        'sort_column'  => 'post_title',
        'hierarchical' => true,
        'exclude'      => '',
        'include'      => '',
        'meta_key'     => '',
        'meta_value'   => '',
        'authors'      => '',
        'child_of'     => 0,
        'parent'       => -1,
        'exclude_tree' => '',
        'number'       => '',
        'offset'       => 0,
        'post_type'    => 'page',
        'post_status'  => 'publish'
    ];
    
    /**
     * Page constructor.
     *
     * @since 0.1
     *
     * @param array $post
     */
    public function __construct( $post = [] ) {
        
        if( $post instanceof WP_Post ) {
            $post = array_change_key_case( $post->to_array(), CASE_LOWER );
        }
    
        parent::__construct( $post );
    }
    
    /**
     * Retrieve the pages child pages
     *
     * @param array $options
     *
     * @since 0.1
     * @return Collection
     */
    public function children( $options = [] ) {
        
        $defaults = array_merge(static::$default_options, [
            'child_of'     => $this->id,
            'post_type'    => static::$post_type
        ]);
        $children = get_pages( array_replace_recursive( $defaults, $options ) );
        $collection = [];
        if( $children !== false ) {
            $collection = array_map( function ( $post ) {
                return static::createFromPost( $post );
            }, $children );
        }
        
        return new Collection($collection);
    }
    
    /**
     * Retrieve ancestors of the page
     *
     * @since 0.1
     * @return Collection
     */
    public function ancestors() {
        return new Collection( array_map( function ( $ancestor ) {
            return static::createFromId( $ancestor );
        }, $this->ancestors ) );
    }
    
    /**
     * Retrieve ancestors of the page
     *
     * @since 0.1
     * @return static
     */
    public function parent() {
        return $this->hasParent() ? static::createFromId( $this->post_parent ) : null;
    }
    
    /**
     * Determine if a page has a parent
     *
     * @since 0.1
     * @return bool
     */
    public function hasParent() {
        return $this->post_parent !== 0;
    }
    
    /**
     * Add chars before the name to emphasise the posts hierarchical position to its ancestors.
     *
     * @since 0.1
     * @return string
     */
    public function hierarchicalName() {
        $text      = $this->post_title;
        $ancestors = $this->post_parent > 0 ? count( $this->ancestors() ) : 0;
        
        return Str::prepend( str_repeat( '— ', $ancestors ), $text );
    }
    
    /**
     * Returns a html string of the pages content
     *
     * @since 0.1
     * @return mixed
     */
    public function content() {
        return $this->post_content;
    }
    
    /**
     * Retrieve page by post
     *
     * @param WP_Post $post
     *
     * @since 0.1
     * @return static
     */
    public static function createFromPost( WP_Post $post ) {
        return new static( array_change_key_case( $post->to_array(), CASE_LOWER ) );
    }
    
    /**
     * Retrieve page by title
     *
     * @param string $title
     *
     * @uses  get_page_by_title()
     * @since 0.1
     * @return static
     */
    public static function createFromTitle( $title ) {
        $post = get_page_by_title( $title );
        return $post ? static::createFromPost($post) : null;
    }
    
    /**
     * Retrieve page by id
     *
     * @param $id
     *
     * @since 0.1
     * @return static
     */
    public static function createFromId( $id ) {
        if( false !== ( $post = WP_Post::get_instance( $id ) ) ) {
            return static::createFromPost( $post );
        }
        return null;
    }
    
    /**
     * Retrieve page by url
     *
     * @param string $url
     *
     * @uses  url_to_postid()
     * @since 0.1
     * @return static
     */
    public static function createFromUrl( $url ) {
        return static::createFromId( url_to_postid( $url ) );
    }
    
    /**
     * Retrieve current page
     *
     * @since 0.1
     * @return static
     */
    public static function current() {
        global $post;
        if( $post ) {
            return static::createFromPost( $post );
        }
        
        return null;
    }
    
    /**
     * Return the pages title
     *
     * @param $prefix
     *
     * @since 0.1
     * @return string
     */
    public function title( $prefix = '' ) {
        $title = $this->post_title;
        
        return Str::notEmpty( $prefix ) ? Str::append( $prefix, $title, ' - ' ) : $title;
    }
    
    /**
     * Add metadata from a "custom field"/Metabox
     *
     * @param string $key
     * @param mixed  $value
     *
     * @since 0.1
     * @return mixed
     */
    public function addMeta( $key, $value ) {
        if( $this->updateAllowed() ) {
            return add_post_meta( $this->id, $key, json_encode( $value, JSON_UNESCAPED_UNICODE ) );
        }
        
        return false;
    }
    
    /**
     * Retrieve the data stored on a specific meta key
     *
     * @param string $key
     * @param mixed  $default
     *
     * @since 0.1
     * @return mixed
     */
    public function getMeta( $key = '', $default = [] ) {
        if( Str::isEmpty( $key ) ) {
            return get_post_meta( $this->id, $key, true );
        }
        $meta = json_decode( get_post_meta( $this->id, $key, true ), true );
        
        return $meta ? $meta : $default;
    }
    
    /**
     * Retrieve the permalink of the page
     *
     * @since 0.1
     * @return false|string
     */
    public function permalink() {
        return get_permalink( $this->id );
    }
    
    /**
     * Retrieve the url to the pages edit page
     *
     * @since 0.1
     * @return false|string
     */
    public function editUrl() {
        return get_edit_post_link( $this->id );
    }
    
    /**
     * Remove the data stored on a specific meta key
     *
     * @param string $key
     *
     * @since 0.1
     * @return mixed
     */
    public function removeMeta( $key ) {
        if( $this->updateAllowed() ) {
            return delete_post_meta( $this->id, $key );
        }
        
        return false;
    }
    
    /**
     * Update the value stored on a specific meta key
     *
     * @param string $key
     * @param mixed  $value
     *
     * @since 0.1
     * @return mixed
     */
    public function updateMeta( $key, $value ) {
        if( $this->updateAllowed() ) {
            return update_post_meta( $this->id, $key, json_encode( $value, JSON_UNESCAPED_UNICODE ) );
        }
        
        return false;
    }
    
    /**
     * Function to handle if a user can update the page
     *
     * @since 0.1
     * @return bool
     */
    public function updateAllowed() {
        if( isset( $_POST[ 'post_type' ] ) && $_POST[ 'post_type' ] === static::$post_type ) {
            return current_user_can( 'edit_' . static::$post_type, $this->id );
        }
    }
}