<?php
/**
 * Twig Extension Class: for handling the environment of Twig
 */
namespace EwTools\Twig;

use Exception;
use Twig_Environment;
use Twig_Lexer;
use Twig_Loader_Chain;
use Twig_LoaderInterface;
use Twig_Parser;
use Twig_SimpleFunction;
use Twig_SourceContextLoaderInterface;

/**
 * Every_Twig
 *
 * @link    http://infomaker.se
 * @package EwTools\Twig
 * @since   EwTools\Twig\Every_Twig 0.1
 */
class Every_Twig {
	
	/**
	 * Contains an instance of the parent for calls to certain functions
	 *
	 * @var \Twig_Environment
	 */
	private static $twig;
	
	/**
	 * Contains an instance of self
	 *
	 * @var Every_Twig
	 */
	private static $instance;
	
	/**
	 * Contains a list of registered loaders
	 *
	 * @var array
	 */
	private static $loaders = [];
	
	/**
	 * Contains whether to use cache
	 *
	 * @var bool
	 */
	protected $cache = false;
	
	/**
	 * Time in seconds to cache the output
	 *
	 * @var int
	 */
	private $cache_time = 60;
	
	/**
	 * Add suffix so we don't have to delete it in deploy every time
	 *
	 * @var string
	 */
	private $cache_key = 'everytwig_compiled_cache_' . GIT_COMMIT;
	
	private function __construct() {}
	
	/**
	 * Create new instance if none exist
	 *
	 * @since 0.1
	 * @return Every_Twig
	 */
	public static function get_instance() {
		if ( null == static::$instance ) {
			static::$instance = new static;
		}
		
		return static::$instance;
	}
    
    /**
     *Initializes the EveryTwig-wrapper and Twig environment
     *
     * @param array $loaders
     * @param array $functions
     * @param array $filters
     * @param array $options
     *
     * @since 0.1
     * @return Twig_Environment
     */
    public function init( array $loaders, array $functions = [], array $filters = [], array $options = ['debug' => false, 'cache' => false] ) {
        
        // Handle options
        $this->handle_options( $options );
        
        foreach ( $loaders as $loader ) {
            $this->add_initial_loader( $loader );
        }
        
//        if( $this->cache ) {
//            if ( false === get_transient( $this->cache_key ) ) {
//                set_transient( $this->cache_key, array(), $this->cache_time );
//            }
//        }
        
        $chain = new Twig_Loader_Chain( static::$loaders );
        $twig  = new Environment( $chain, $options );
        
        $twig->addFunction( $this->default_php_function() );
        
        foreach ( $functions as $function ) {
            $twig->addFunction( $function );
        }
        
        foreach ( $filters as $filter ) {
            $twig->addFilter( $filter );
        }
        
        $twig->setLexer( new Twig_Lexer( $twig ) );
        $twig->setParser( new Twig_Parser( $twig ) );
        
        return static::$twig = $twig;
    }
    
    /**
     * Helper function to add a Twig_LoaderInterface loader on init
     *
     * @param Twig_LoaderInterface|Twig_SourceContextLoaderInterface $loader
     *
     * @return void
     * @since 0.1
     */
    private function add_initial_loader( Twig_SourceContextLoaderInterface $loader ) {
        array_push( static::$loaders, $loader );
    }
    
    /**
     * Function to add loaders after a Twig environment has been initialized
     *
     * @param Twig_LoaderInterface|Twig_SourceContextLoaderInterface $loader
     *
     * @throws Exception
     * @since 0.1
     */
    public function add_loader( Twig_SourceContextLoaderInterface $loader ) {
        if( is_null(static::$twig) ) {
            throw new Exception('Function init needs to be run in order to add extra loaders');
        }
        
        array_push(static::$loaders, $loader);
        static::$twig->setLoader( new Twig_Loader_Chain(static::$loaders) );
    }
	
	/**
	 * Handles the specific cache options and trims them away from the options array.
	 *
	 * @param array $options
	 *
	 * @since 0.1
	 * @return void
	 */
	private function handle_options( &$options ) {
        $options['cache'] = true;
	    
		if ( isset( $options['cache'] ) ) {
			$this->cache = $options['cache'];
			unset( $options['cache'] );
		}
		
		if ( isset( $options['cachetime'] ) && intval( $options['cachetime'] ) > 0 ) {
			$this->cache_time = intval( $options['cachetime'] );
			unset( $options['cachetime'] );
		}
		// Add custom cache system to Twig
        $options['cache'] = $this->cache ? ObjectCacheSystem::setup($this->cache_key, $this->cache_time) : false;
	}
	
	/**
	 * Twig_SimpleFunction to be able to call default php functions from within twig templates
	 *
	 * @since 0.1
	 * @return Twig_SimpleFunction
	 */
	protected function default_php_function() {
		$phpFunction = new Twig_SimpleFunction( 'php_function', function( $name, $args = [] ) {
			if (! function_exists( $name ) ) {
                throw new Exception('Function dose not exist');
			}
			
			if ( strpos( $name, '::' ) ) {
				$static_method = explode( '::', $name );
				if (! method_exists( $static_method[0], $static_method[1] ) ) {
                    throw new Exception('Method '. $static_method[1] .' can not be found on ' . $static_method[0]);
				}
			}
			
            return call_user_func_array( $name, $args );
			
		} );
		
		return $phpFunction;
	}
	
	/**
	 * Render function of EveryTwig, loads a specific template and works with cached compiled templates.
	 *
	 * @param string $template
	 * @param array  $context
	 *
	 * @since 0.1
	 * @return void
	 */
	public function render( $template, array $context = [] ) {
        static::$twig->display($template, $context );
	}
	
	/**
	 * Return the view as rendered HTML by passed template
	 *
	 * @param string $template
	 * @param array  $context
	 *
	 * @since 0.1
	 * @return mixed
	 */
	public function view( $template, array $context = [] ) {
		return static::$twig->render( $template, $context );
	}
	
	/**
	 * Loads a specific template and compiles it. If cache is on, save the compiled class in cac3he to be evaluated
	 * later on.
	 *
	 * @param string   $name
	 * @param int|null $index
	 *
	 * @since 0.1
	 * @return mixed
	 */
//	public function loadTemplate( $name, $index = null ) {
//		$cls = static::$twig->getTemplateClass( $name, $index );
//
//		if ( isset( static::$twig->loadedTemplates[ $cls ] ) ) {
//			return static::$twig->loadedTemplates[ $cls ];
//		}
//
//		if ( ! class_exists( $cls, false ) ) {
//			if ( false === ( $cache = $this->get_cached_template( $cls ) ) ) {
//				$source = static::$twig->compileSource( static::$twig->getLoader()->getSource( $name ), $name );
/*				eval( '?>' . $source );*/
//
//				if ( $this->cache ) {
//					$this->cache_template( $cls, serialize( $source ) );
//				}
//			} else {
/*				eval( '?>' . unserialize( $cache ) );*/
//			}
//		}
//
//		if ( ! static::$twig->runtimeInitialized ) {
//			static::$twig->initRuntime();
//		}
//		return static::$twig->loadedTemplates[ $cls ] = new $cls( static::$twig );
//	}
	
	/**
	 * Helper function to fetch cached template
	 *
	 * @param string $name
	 *
	 * @since 0.1
	 * @return bool
	 */
//	private function get_cached_template( $name ) {
//		if ( ! $this->cache ) {
//			return false;
//		}
//
//		$cache = get_transient( $this->cache_key );
//
//		return $cache !== false && isset( $cache[ $name ] ) ? $cache[ $name ] : false;
//	}
	
	/**
	 * Helper function to cache template
	 *
	 * @param string $name
	 * @param mixed  $data
	 *
	 * @since 0.1
	 * @return void
	 */
//	private function cache_template( $name, $data ) {
//		$cache          = get_transient( $this->cache_key );
//		$cache[ $name ] = $data;
//		set_transient( $this->cache_key, $cache, $this->cache_time );
//	}
    
    /**
     * Registers a Global.
     *
     * New globals can be added before compiling or rendering a template;
     * but after, you can only update existing globals.
     *
     * @param string $name  The global name
     * @param mixed  $value The global value
     */
//    public function addGlobal($name, $value) {
//        static::$twig->addGlobal($name, $value);
//	}
	/**
	 * Forward to local instance::getTokenParsers()
	 * @see   Twig_Environment::getTokenParsers()
	 *
	 * @since 0.1
	 * @return \Twig_TokenParserBrokerInterface
	 */
//	public function getTokenParsers() {
//		return static::$twig->getTokenParsers();
//	}
	
	/**
	 * Forward to local instance::getNodeVisitors()
	 * @see   Twig_Environment::getNodeVisitors()
	 *
	 * @since 0.1
	 * @return \Twig_NodeVisitorInterface[]
	 */
//	public function getNodeVisitors() {
//		return static::$twig->getNodeVisitors();
//	}
	
	/**
	 * Forward to local instance::getUnaryOperators()
	 * @see   Twig_Environment::getUnaryOperators()
	 *
	 * @since 0.1
	 * @return array
	 */
//	public function getUnaryOperators() {
//		return static::$twig->getUnaryOperators();
//	}
	
	/**
	 * Forward to local instance::getBinaryOperators()
	 * @see   Twig_Environment::getBinaryOperators()
	 *
	 * @since 0.1
	 * @return array
	 */
//	public function getBinaryOperators() {
//		return static::$twig->getBinaryOperators();
//	}
	
	/**
	 * Forward to local instance::computeAlternatives()
	 * @see   Twig_Environment::computeAlternatives()
	 *
	 * @param $name
	 * @param $items
	 *
	 * @since 0.1
	 * @return array
	 */
//	public function computeAlternatives( $name, $items ) {
//		return static::$twig->computeAlternatives( $name, $items );
//	}
	
	/**
	 * Forward to local instance::getTags()
	 * @see   Twig_Environment::getTags()
	 *
	 * @since 0.1
	 * @return \Twig_TokenParserInterface[]
	 */
//	public function getTags() {
//		return static::$twig->getTags();
//	}
	
	/**
	 * Forward to local instance::getFunction()
	 * @see   Twig_Environment::getFunction()
	 *
	 * @param string $name
	 *
	 * @since 0.1
	 * @return false|\Twig_Function
	 */
//	public function getFunction( $name ) {
//		return static::$twig->getFunction( $name );
//	}
	
	/**
	 * Forward to local instance::getFunctions()
	 * @see   Twig_Environment::getFunctions()
	 *
	 * @since 0.1
	 * @return \Twig_FunctionInterface[]
	 */
//	public function getFunctions() {
//		return static::$twig->getFunctions();
//	}
	
	/**
	 * Forward to local instance::getBaseTemplateClass()
	 * @see   Twig_Environment::getBaseTemplateClass()
	 *
	 * @since 0.1
	 * @return string
	 */
//	public function getBaseTemplateClass() {
//		return static::$twig->getBaseTemplateClass();
//	}
	
	/**
	 * Forward to local instance::getExtension()
	 * @see   Twig_Environment::getExtension()
	 *
	 * @param string $name
	 *
	 * @since 0.1
	 * @return \Twig_ExtensionInterface
	 */
//	public function getExtension( $name ) {
//		return static::$twig->getExtension( $name );
//	}
	
	/**
	 * Forward to local instance::getExtensions()
	 * @see   Twig_Environment::getExtensions()
	 *
	 * @since 0.1
	 * @return \Twig_ExtensionInterface[]
	 */
//	public function getExtensions() {
//		return static::$twig->getExtensions();
//	}
	
	/**
	 * Forward to local instance::getFilter()
	 * @see   Twig_Environment::getFilter()
	 *
	 * @param string $name
	 *
	 * @since 0.1
	 * @return false|\Twig_Filter
	 */
//	public function getFilter( $name ) {
//		return static::$twig->getFilter( $name );
//	}
	
	/**
	 * Forward to local instance::getFilters()
	 * @see   Twig_Environment::getFilters()
	 *
	 * @since 0.1
	 * @return \Twig_FilterInterface[]
	 */
//	public function getFilters() {
//		return static::$twig->getFilters();
//	}
}